(module info setup/infotab (#%module-begin (define collection "preprocessor") (define mzscheme-launcher-names (quote ("mzpp" "mztext"))) (define mzscheme-launcher-libraries (quote ("mzpp-run.rkt" "mztext-run.rkt"))) (define scribblings (quote (("scribblings/preprocessor.scrbl" (multi-page) (legacy))))) (define deps (quote ("scheme-lib" "base" "compatibility-lib"))) (define build-deps (quote ("racket-doc" "scribble-lib"))) (define pkg-desc "Preprocessors for text with embedded Racket code (mostly replaced by scribble/text)") (define pkg-authors (quote (eli))) (define license (quote (Apache-2.0 OR MIT)))))
