(module info setup/infotab (#%module-begin (define collection (quote multi)) (define deps (quote ("images-lib" "images-gui-lib" "images-doc"))) (define implies (quote ("images-lib" "images-gui-lib" "images-doc"))) (define pkg-desc "Functions (and docs and tests) for constructing icons and logos") (define pkg-authors (quote (ntoronto))) (define license (quote (Apache-2.0 OR MIT)))))
