(module info setup/infotab (#%module-begin (define collection (quote multi)) (define deps (quote ("compatibility-lib" "compatibility-doc"))) (define implies (quote ("compatibility-lib" "compatibility-doc"))) (define pkg-desc "Libraries that implement legacy interfaces") (define pkg-authors (quote (eli mflatt robby samth))) (define license (quote (Apache-2.0 OR MIT)))))
