(module info setup/infotab (#%module-begin (define collection "ds-store") (define deps (quote ("ds-store-lib" "ds-store-doc" "base"))) (define implies (quote ("ds-store-lib" "ds-store-doc"))) (define pkg-desc "Libraries for manipulating \".DS_Store\" files") (define pkg-authors (quote (mflatt))) (define license (quote (Apache-2.0 OR MIT)))))
