;; This template is turned into a machine-specific ".def" file
;; by the `machine.def` script

(define-constant machine-type (constant machine-type-$(M)))
(features iconv expeditor)
(define-constant time-t-bits $(Mtimet))
(include "$(March).def")
(include "default.def")
