(module info setup/infotab (#%module-begin (define collection (quote multi)) (define deps (quote ("base" ("scribble-lib" #:version "1.46")))) (define build-deps (quote ("scheme-lib" "at-exp-lib"))) (define pkg-desc "Racket Documentation driver") (define pkg-authors (quote (eli jay matthias mflatt robby ryanc samth))) (define version "1.3") (define binary-keep-files (quote ("scribblings" "scribblings/main/compiled/acks_scrbl.zo" "scribblings/main/compiled/license_scrbl.zo" "scribblings/main/compiled/local-redirect_scrbl.zo" "scribblings/main/compiled/release_scrbl.zo" "scribblings/main/compiled/search_scrbl.zo" "scribblings/main/compiled/start_scrbl.zo" "scribblings/main/user/compiled/local-redirect_scrbl.zo" "scribblings/main/user/compiled/release_scrbl.zo" "scribblings/main/user/compiled/search_scrbl.zo" "scribblings/main/user/compiled/start_scrbl.zo"))) (define license (quote (Apache-2.0 OR MIT)))))
