(module info setup/infotab (#%module-begin (define collection (quote multi)) (define deps (quote (("base" #:version "6.8.0.2") "data-lib" "math-lib"))) (define build-deps (quote ("rackunit-lib"))) (define pkg-desc "implementation (no documentation) of \"data/enumerate\"") (define pkg-authors (quote (maxsnew jay robby))) (define version "1.3") (define license (quote (Apache-2.0 OR MIT)))))
