(module info setup/infotab (#%module-begin (define collection (quote multi)) (define deps (quote ("htdp-lib" "htdp-doc"))) (define implies (quote ("htdp-lib" "htdp-doc"))) (define pkg-desc "Teaching languages for _How to Design Programs_") (define pkg-authors (quote (matthias mflatt robby))) (define license (quote (Apache-2.0 OR MIT)))))
