(module info setup/infotab (#%module-begin (define collection (quote multi)) (define build-deps (quote ("racket-index" "net-doc" "scheme-lib" "draw-doc" "at-exp-lib" "base" "compatibility-lib" "draw-lib" "pict-lib" "pict-doc" "sandbox-lib" "scribble-lib" "scribble-text-lib" "racket-doc"))) (define update-implies (quote ("scribble-lib"))) (define pkg-desc "documentation part of \"scribble\"") (define pkg-authors (quote (mflatt eli))) (define license (quote (Apache-2.0 OR MIT)))))
