#lang info
;; skip this file because we want to preserve the source
;; location information in the syntax templates so the
;; redex examples typset properly
(define compile-omit-paths '("typesetting-and-macros.scrbl"))

