(module info setup/infotab (#%module-begin (define collection (quote multi)) (define deps (quote ("string-constants-lib" "base"))) (define build-deps (quote ("racket-doc" "scribble-lib"))) (define pkg-desc "String constants documentation") (define update-implies (quote ("string-constants-lib"))) (define pkg-authors (quote (robby))) (define license (quote (Apache-2.0 OR MIT)))))
