(module info setup/infotab (#%module-begin (define collection "sgl") (define compile-omit-paths (quote ("examples"))) (define test-omit-paths (quote ("examples"))) (define scribblings (quote (("scribblings/sgl.scrbl" (multi-page) (gui-library))))) (define deps (quote ("scheme-lib" "base" "compatibility-lib" "gui-lib"))) (define build-deps (quote ("draw-doc" "gui-doc" "racket-doc" "scribble-lib"))) (define pkg-desc "Legacy OpenGL library") (define pkg-authors (quote (jay))) (define license (quote (Apache-2.0 OR MIT)))))
