(module info setup/infotab (#%module-begin (define collection "racket-cheat") (define deps (quote ("base" "scribble-lib"))) (define build-deps (quote ("db-doc" "db-lib" "drracket" "net-doc" "net-lib" "parser-tools-doc" "parser-tools-lib" "pict-doc" "pict-lib" "racket-doc" "sandbox-lib" "slideshow-doc" "slideshow-lib"))) (define scribblings (quote (("racket-cheat.scrbl" () (getting-started))))) (define pkg-desc "a cheat sheet for Racket") (define pkg-authors (quote (jay))) (define license (quote (Apache-2.0 OR MIT)))))
