(module info setup/infotab (#%module-begin (define collection (quote multi)) (define deps (quote (("base" #:version "6.2.900.10") "scribble-lib" ("string-constants-lib" #:version "1.12") "racket-index"))) (define build-deps (quote ("at-exp-lib" "rackunit-lib"))) (define pkg-desc "Non-GUI code implementing programmatic interfaces to some IDE tools that DrRacket supports") (define pkg-authors (quote (robby))) (define version "1.3") (define license (quote (Apache-2.0 OR MIT)))))
