#lang racket/base
(require rackunit/docs-complete)
(check-docs (quote games/show-scribbling))
(check-docs (quote games/show-help))
(check-docs (quote games))
(check-docs (quote games/gl-board-game))
(check-docs (quote games/cards))
