(module info setup/infotab (#%module-begin (define collection (quote multi)) (define deps (quote ("base" "compiler-lib" "rackunit-lib" "redex-lib" "redex-examples" "math-lib" "plot-lib"))) (define build-deps (quote ())) (define pkg-desc "PLT Redex Benchmark") (define pkg-authors (quote (robby bfetscher))) (define license (quote (Apache-2.0 OR MIT)))))
