#lang info

(define test-omit-paths '("example-min.rkt"
                          "example.rkt"
                          "minimal.rkt"
                          "more-tests-margin-aligned.rkt"
                          "more-tests-margin.rkt"
                          "more-tests-min-stretchable.rkt"
                          "more-tests-min.rkt"
                          "more-tests-text.rkt"
                          "more-tests.rkt"
                          "old-bugs"
                          "stretchable-editor-snip-test-min.rkt"
                          "stretchable-editor-snip-test.rkt"
                          "test-locked-pasteboard.rkt"))
