(module info setup/infotab (#%module-begin (define collection (quote multi)) (define deps (quote ("base" "rackunit-lib"))) (define pkg-desc "Testing framework") (define pkg-authors (quote (eli))) (define license (quote (Apache-2.0 OR MIT)))))
